# Raymond Tree Algorithm
# Test user input
# n = 5

# request_queue = {0: [], 1: [], 2: [], 3: [], 4: []}
# holder = {0: 0, 1: 0, 2: 0, 3: 1, 4: 1}
# token = {0: 1, 1: 0, 2: 0, 3: 0, 4: 0}

# adj_matrix = [[1, 0, 0, 0, 0],
#               [1, 0, 0, 0, 0],
#               [1, 0, 0, 0, 0],
#               [0, 1, 0, 0, 0],
#               [0, 1, 0, 0, 0]]

# Number of nodes
n = 5

# Initializing data structures
request_queue = {i: [] for i in range(n)}
holder = {i: i for i in range(n)}
token = {i: 1 if i == 0 else 0 for i in range(n)}

adj_matrix = [[1, 0, 0, 0, 0],
              [1, 0, 0, 0, 0],
              [1, 0, 0, 0, 0],
              [0, 1, 0, 0, 0],
              [0, 1, 0, 0, 0]]

print("\nRaymond Tree Based Mutual Exclusion")
print("\nAdjacency Matrix for spanning tree:\n")

for row in adj_matrix:
    print(*row)

# Function to find the parent of a process
def find_parent(req_process):
    # Sending the request to the root that has the token
    request_queue[req_process].append(req_process)

    for i in range(n):
        if adj_matrix[req_process][i] == 1:
            parent = i
            request_queue[parent].append(req_process)
            break

    print("\nProcess {} sending request to Parent Process {}".format(
        req_process, parent))
    print("Request queue: ", request_queue)

    # Check if parent has the token or not
    if token[parent] == 1:
        return parent
    else:
        parent = find_parent(parent)

    return parent


# Input process that wants to enter CS
while True:
    # Example: Enter the process who wants to enter CS (0-4): 2
    req_process = int(input("\nEnter the process who wants to enter CS (0-{}): ".format(n - 1)))
    if 0 <= req_process < n:
        break
    else:
        print("Invalid process ID. Please enter a valid process ID.")

parent = find_parent(req_process)

while token[req_process] != 1:
    child = request_queue[parent][0]
    request_queue[parent].pop(0)

    holder[parent] = child
    holder[child] = parent
    token[parent] = 0
    token[child] = 1

    print("\nParent process {} has the token and sends the token to the request process {}".format(
        parent, child))
    print("Request Queue: ", request_queue)
    parent = child

if token[parent] == 1 and request_queue[parent][0] == parent:
    request_queue[parent].pop(0)
    holder[parent] = parent
    print("\nProcess {} enters the Critical Section".format(parent))
    print("Request Queue: ", request_queue)

if len(request_queue[parent]) == 0:
    print("\nRequest Queue of process {} is empty. Therefore Release Critical Section".format(parent))

print("\nHolder: ", holder)



import socket

# Set up client
client = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
print("\n", client, "\n")
client.connect(('127.0.0.1', 8888))

# Send data to server
message = b"Hello, server!"
sent = client.send(message)
print("\n", client, "\n")

# Check if message has been sent
if sent == len(message):
    print("Message sent successfully!")
else:
    print("Message not sent completely.")

# Receive data from server
response = client.recv(1024)
print("\n", client, "\n")
print("\nResponse: ", response.decode(), "\n")

# Close client connection
client.close()



def print_server_load(num_servers, num_processes):
    processes_per_server = num_processes // num_servers
    extra_processes = num_processes % num_servers

    for i in range(extra_processes):
        print(f"Server {i + 1} has {processes_per_server + 1} processes")

    for i in range(extra_processes, num_servers):
        print(f"Server {i + 1} has {processes_per_server} processes")


def display_menu():
    print("1. Add Server")
    print("2. Remove Server")
    print("3. Add Processes")
    print("4. Remove Processes")
    print("5. Exit")


def main():
    num_servers = int(input("Enter number of servers: "))
    num_processes = int(input("Enter number of processes: "))

    while True:
        print_server_load(num_servers, num_processes)
        display_menu()
        choice = int(input("> "))

        if choice == 1:
            num_servers += int(input("Enter number of servers to be added: "))
        elif choice == 2:
            num_servers -= int(input("Enter number of servers to be removed: "))
        elif choice == 3:
            num_processes += int(input("Enter number of processes to be added: "))
        elif choice == 4:
            num_processes -= int(input("Enter number of processes to be removed: "))
        elif choice == 5:
            return
        else:
            print("Invalid choice. Please try again.")


if __name__ == "__main__":
    main()



n = int(input('Enter number of nodes : '))
faulty = int(input(f'Enter faulty node (out of {1} to {n}) : '))
detector = int(input('Enter node that detected failure first : '))


def find_coordinator(detector, faulty, n):
    for node in range(detector + 1, n + 1):  # Start from the next node after the detector
        if node == faulty:
            continue
        print()
        print(f'---- Node {node} sending ELECTION message ----')
        oks = 0
        for neighbor in range(node + 1, n + 1):
            if neighbor != faulty:
                oks += 1
            print(f'ELECTION message sent to Node {neighbor}')

        if oks > 0:
            print(f'{oks} (OKs) received by Node {node}')
        else:
            print('Active higher priority process does NOT exist..')
            return node



def bully(detector, faulty, n):
    print(f'The Coordinator (Node {faulty}) has failed...')
    print(f'Node {detector} detected failure of coordinator...')
    print(f'Node {detector} initiating election process...')

    new_coordinator = find_coordinator(detector, faulty, n)

    print()
    print('----- RESULT OF ELECTION PROCESS-----')
    print(f'Node {new_coordinator} elected as new coordinator !!')
    print(
        f'Node {new_coordinator} sending message to inform that it is elected as new coordinator...')
    for neighbor in range(1, new_coordinator):
        print(f'message sent to Node {neighbor}')


bully(detector, faulty, n)

'''
About program
The program takes three inputs: the number of nodes in the system, the node that detected the failure first, and the faulty node. Then it simulates the election process to select a new coordinator.

The find_coordinator() function is responsible for sending ELECTION messages to all the nodes with a higher ID than the current node. If any of the higher nodes respond with an OK message, the current node cancels the election process and waits for the higher node to become the coordinator.

If no higher node responds, the current node declares itself as the coordinator and sends a COORDINATOR message to all the lower nodes.

The bully() function initiates the election process and calls the find_coordinator() function. Once the new coordinator is selected, the function prints the result and sends a message to inform all the lower nodes about the new coordinator.

About code
The program simulates the Bully Election Algorithm in a network of three nodes. The user is prompted to input the number of nodes, the index of the faulty node, and the index of the node that detects the failure first.

In the given scenario, there are three nodes, and Node 2 is the faulty node that has failed. Node 1 detects the failure first and initiates the election process.

Node 1 sends an election message to nodes 2 and 3, and both nodes respond with an OK message, indicating that they are still active. Since Node 3 has a higher priority than Node 2, it sends its own election message to Node 2, which does not respond since it has already failed. Thus, Node 3 declares itself as the new coordinator and sends a message to inform nodes 1 and 2 of its election.

The output shows the messages sent and received during the election process, and the final result of the election, which declares Node 3 as the new coordinator.
'''



import socket

# Set up server
server = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
server.bind(('127.0.0.1', 8888))
server.listen(5)
print("\n Server : ", server, "\n" )

print("Server listening on port 8888")

while True:
    # Accept connection and handle data
    client, _ = server.accept()

    print("\n Client : ", client, "\n" )

    data = client.recv(1024)
    print("Received:", data.decode())
    client.send(data)
    client.close()



def display(mx, alloc, need, avail):
    print('Available Resources: ', avail)
    print('Allocated Matrix:')
    for row in alloc:
        print(*row)
    print()
    print('Max Matrix:')
    for row in mx:
        print(*row)
    print()
    print('Need Matrix:')
    for row in need:
        print(*row)
    print()


def bankers(n, r, mx, alloc, need, avail):
    print('***** Initially *****')
    display(mx, alloc, need, avail)

    safe_sequence = []

    while True:
        found = False

        for i in range(n):
            if i + 1 in safe_sequence:
                continue

            can_allocate = True
            for j in range(r):
                if need[i][j] > avail[j]:
                    can_allocate = False
                    break

            if can_allocate:
                safe_sequence.append(i + 1)
                for j in range(r):
                    avail[j] += alloc[i][j]
                    alloc[i][j] = 0
                    need[i][j] = '-'  # Mark as done

                print(f'***** After allocating resources to P{i + 1} *****')
                print(f'P{i + 1} can be allocated resources for execution..')
                display(mx, alloc, need, avail)
                print()
                found = True
                break

        if not found:
            print("System is NOT in a safe state !!")
            break

        if len(safe_sequence) == n:
            print("System is in a safe state !!")
            print("Safe Sequence is:", end=" ")
            for i in safe_sequence:
                print(f"P{i}", end=" ")
            print()
            break


n = 5
r = 3
alloc = [[0, 1, 0], [2, 0, 0], [3, 0, 2], [2, 1, 1], [0, 0, 2]]
mx = [[7, 5, 3], [3, 2, 2], [9, 0, 2], [2, 2, 2], [4, 3, 3]]
need = [[7, 4, 3], [1, 2, 2], [6, 0, 0], [0, 1, 1], [4, 3, 1]]
avail = [3, 3, 2]

bankers(n, r, mx, alloc, need, avail)



import socket
import threading

# Function to receive messages from server
def receive_messages(client_socket):
    while True:
        message = client_socket.recv(1024).decode()
        print(message)

# Set up client
client = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
client.connect(('127.0.0.1', 8888))

# Start a thread to receive messages from server
receive_thread = threading.Thread(target=receive_messages, args=(client,))
receive_thread.start()

# Main thread to send messages to server
while True:
    message = input()
    client.send(message.encode())



# Distributed Global Averaging

import threading

# Take number of nodes as user input
NUM_NODES = int(input("Enter the number of nodes: "))

# Initialize local values for each node
local_values = []
for i in range(NUM_NODES):
    value = int(input(f"Enter local value for node {i + 1}: "))
    local_values.append(value)

global_average = 0

# Lock for synchronization
lock = threading.Lock()

# Function to compute the average of local values
def compute_local_average():
    with lock:
        local_average = sum(local_values) / len(local_values)
        print(f"Local average: {local_average}\n")
        return local_average

# Function to update global average using local averages
def update_global_average():
    global global_average
    threshold = 1e-6  # Convergence threshold
    while True:
        local_average = compute_local_average()
        with lock:
            global_average = (global_average + local_average) / 2
        print(f"Updated global average: {global_average}\n")
        if abs(global_average - local_average) < threshold:
            break

# Start a thread to continuously update the global average
update_thread = threading.Thread(target=update_global_average)
update_thread.start()

# Main thread waits for the update thread to finish
update_thread.join()

print(f"Final global average: {global_average}")



import socket
import threading

# Function to handle each client connection
def handle_client(client_socket, clients):
    while True:
        # Receive message from client
        message = client_socket.recv(1024).decode()
        if not message:  # If message is empty, client has disconnected
            break

        # Broadcast message to all clients
        for client in clients:
            if client != client_socket:  # Don't send the message back to the sender
                client.send(message.encode())

    # Remove client from list after they disconnect
    clients.remove(client_socket)
    client_socket.close()

# Set up server
server = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
server.bind(('127.0.0.1', 8888))
server.listen(5)

print("Server listening on port 8888")

# List to store client sockets
clients = []

# Accept incoming connections and handle them in separate threads
while True:
    client_socket, _ = server.accept()  # Accept connection
    clients.append(client_socket)  # Add client socket to list
    client_handler = threading.Thread(target=handle_client, args=(client_socket, clients))  # Start thread to handle client
    client_handler.start()



def main():
    num_processes = int(input("Enter the number of processes: "))
    num_cs = int(input("Enter the number of processes that want to enter the critical section: "))

    # Dictionary to store process IDs and their corresponding timestamps
    process_map = {}
    timestamps = []

    # Input process IDs and timestamps
    for _ in range(num_cs):
        process, timestamp = input("Enter the Process ID and Timestamp of process (e.g., '2 10'): ").split()
        process_map[int(timestamp)] = int(process)
        timestamps.append(int(timestamp))

    print()  # For better output formatting

    # Sort timestamps to ensure requests are processed in order
    timestamps.sort()

    # Simulate process requests
    for time in timestamps:
        process_cs = process_map[time]
        for i in range(num_processes):
            if process_cs != i:
                print(f"Process {process_cs} has requested Process {i}")
        print()  # Empty line for better output formatting

    # Simulate process acceptances and entry to critical section
    for time in timestamps:
        process_cs = process_map[time]
        for i in range(num_processes):
            if process_cs != i:
                print(f"Process {i} has accepted the request of process {process_cs}")
        print()  # Empty line for better output formatting
        print(f'Process {process_cs} has now entered the critical section')
        print(f'Process {process_cs} has now exited the critical section')
        print()

if __name__ == "__main__":
    main()



import socket
import threading

SERVER_HOST = "127.0.0.1"
SERVER_PORT = 1234

class ClientThread(threading.Thread):
    def __init__(self, name):
        super().__init__()
        self.name = name

    def run(self):
        try:
            # Connect to server
            with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as client_socket:
                client_socket.connect((SERVER_HOST, SERVER_PORT))

                # Send message to server
                message = f"{self.name} is sending a message"
                client_socket.sendall(message.encode())

                # Receive response from server
                response = client_socket.recv(1024).decode()
                print(f"{self.name} received response: {response}")

        except socket.error as e:
            print(f"Socket error: {e}")

def main():
    # Create and start client threads
    threads = []
    for i in range(5):
        thread = ClientThread(f"Thread {i}")
        thread.start()
        threads.append(thread)

    # Wait for all threads to complete
    for thread in threads:
        thread.join()

if __name__ == "__main__":
    main()



import socket
import threading

SERVER_HOST = "127.0.0.1"
SERVER_PORT = 1234

def handle_client(client_socket, address):
    try:
        # Receive message from client
        message = client_socket.recv(1024).decode()
        print(f"Received message from {address}: {message}")

        # Send response back to client
        response = f"Response to {message}"
        client_socket.sendall(response.encode())
    except socket.error as e:
        print(f"Socket error: {e}")
    finally:
        client_socket.close()

def main():
    try:
        # Create server socket
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as server_socket:
            # Bind server socket to address
            server_socket.bind((SERVER_HOST, SERVER_PORT))
            print("Server started")

            # Listen for incoming connections
            server_socket.listen(5)

            while True:
                # Accept client connection
                client_socket, address = server_socket.accept()
                print(f"Client connected: {address}")

                # Create and start client handler thread
                client_handler = threading.Thread(target=handle_client, args=(client_socket, address))
                client_handler.start()

    except socket.error as e:
        print(f"Socket error: {e}")

if __name__ == "__main__":
    main()



# Code to synchronize time

import time
import random

# Time server node index
TIME_SERVER_NODE = 0

# Node class representing a node in the network
class Node:
    def __init__(self, id, offset):
        self.id = id
        self.offset = offset
        # Initialize the clock with the current time in seconds
        self.clock = time.time() + random.uniform(-1, 1)  # Add a random offset

    # Method to adjust the clock offset of the node
    def adjust_clock(self, offset):
        self.offset += offset
        self.clock += offset  # Update the clock value

# Function to synchronize all nodes in the network
def synchronize_clocks(nodes):
    # Calculate the time of the time server node
    server_time = nodes[TIME_SERVER_NODE].clock
    
    # Adjust offset to each node
    for node in nodes:
        if node.id != TIME_SERVER_NODE:
            node_time = node.clock
            offset = server_time - node_time
            node.adjust_clock(offset)

# Main function
def main():
    # Take input for the number of nodes
    num_nodes = int(input("Enter the number of nodes: "))

    # Create nodes in the network with user-defined offsets
    nodes = []
    for i in range(num_nodes):
        offset = float(input(f"Enter the offset for node {i + 1} (in seconds): "))
        nodes.append(Node(i + 1, offset))

    # Print initial clock values of all nodes
    print("\nBefore synchronization:")
    for node in nodes:
        print(f"Node {node.id} clock: {node.clock}")

    return nodes

if __name__ == "__main__":
    nodes = main()

    # Synchronize clocks of all nodes after the main process completes
    synchronize_clocks(nodes)

    # Print clock values of all nodes after synchronization
    print("\nAfter synchronization:")
    for node in nodes:
        print(f"Node {node.id} clock: {node.clock}")



from setuptools import setup, find_packages

VERSION = '4.2.9'
DESCRIPTION = 'Helps users code efficently'
LONG_DESCRIPTION = 'A package that allows to build simple codes in a well defined and efficient manner.'

# Setting up
setup(
    name="skleanr",
    version=VERSION,
    author="Tom Hanks",
    author_email="<tomhanks02@gmail.com>",
    description=DESCRIPTION,
    long_description_content_type="text/markdown",
    packages=find_packages(),
    keywords=['python', 'video', 'stream', 'video stream', 'camera stream', 'sockets'],
    classifiers=[
        "Development Status :: 1 - Planning",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3",
        "Operating System :: Unix",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
    ]
)


